# icsdll/tests/conftest.py

import os
import sys

if os.environ.get('VSCODE_CWD'):
    # work around pytest not using PYTHONPATH in VSCode
    sys.path.insert(
        0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    )


def pytest_report_header(config):
    try:
        from numpy import __version__ as numpy

        from icsdll import __version__
    except Exception:
        pass
    else:
        return f'versions: icsdll-{__version__}, numpy-{numpy}'


collect_ignore = ['_tmp', 'data']



# icsdll.py

# Copyright (c) 2016-2026, Christoph Gohlke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Interface to the image correlation spectroscopy library ICSx64.dll.

ICSdll is a Python ctypes interface to the Image Correlation Spectroscopy
Dynamic Link Library (ICSx64.dll) developed at the Laboratory for Fluorescence
Dynamics (LFD) for the Globals for Images SimFCS software.

ICSx64.dll is implemented in C++ using the Intel(r) oneAPI Math Kernel Library
and OpenMP. It provides functions and classes for the analysis of fluorescence
time series data:

- 1D, 2D, and 3D auto- and cross-correlation
- Image pair correlation function (ipCF)
- Airy detector pair correlation function (apCF)
- Image mean square displacement (iMSD)
- Line spatio-temporal image correlation spectroscopy (lSTICS)
- Fit 1D pair correlation functions to the results of ipCF analysis
- Subtract immobile fractions
- Correct photo-bleaching
- 1D DFTs of image stack

:Author: `Christoph Gohlke <https://www.cgohlke.com>`_
:License: BSD-3-Clause
:Version: 2026.3.3

Requirements
------------

This revision was tested with the following requirements and dependencies
(other versions may work):

- `CPython <https://www.python.org>`_ 3.11.9, 3.12.10, 3.13.12, 3.14.3 64-bit
- `Numpy <https://pypi.org/project/numpy>`_ 2.4.2
- `Intel(r) oneAPI Math Kernel Library <https://software.intel.com/mkl>`_
  2025.3 (build)
- `Visual Studio 2022 C++ compiler <https://visualstudio.microsoft.com/>`_
  (build)

Revisions
---------

2026.3.3

- Rename class xyt to yxt to match the C API (breaking).
- Enforce keyword-only arguments (breaking).
- Fix code review issues.
- Improve documentation.
- Rebuild package with oneAPI MKL 2025.3.

2025.8.8

- Support Python 3.14.
- Rebuild package with oneAPI MKL 2025.2.

2025.1.6

- Support Python 3.13 and numpy 2.
- Rebuild package with oneAPI MKL 2025.0.

2024.1.6

- Rebuild package with oneAPI MKL 2024.0.0.

2023.1.6

- Rebuild package with oneAPI MKL 2022.2.1 and Visual Studio 2022.
- Update metadata.

2022.2.6

- Rebuild package with oneAPI 2022.

2021.3.2

- Rebuild package.

2019.11.22

- Wrap yxt_dft functions.
- Upgrade to ICSx64.DLL version 2019.11.22.

2019.7.10

- Pass 22 tests.
- Wrap apcf, imsd, and lstics functions.
- Raise IcsError in case of DLL function errors.
- Use ICSx64.DLL version 2019.7.10.

2019.5.22

- Initial release based on ICSx64.DLL version 2018.7.18.

Notes
-----

ICSdll was developed, built, and tested on 64-bit Windows only.

The API is not stable and might change between revisions.

Refer to the C++ header file and source code for function signatures.

References
----------

1. `ipcf.ipynb <https://github.com/cgohlke/ipcf.ipynb>`_,
   Pair correlation function analysis of fluorescence fluctuations in
   big image time series using Python.
2. `Globals for Images SimFCS <https://www.lfd.uci.edu/globals/>`_,
   Software for fluorescence image acquisition, analysis, and simulation.
3. `Globals for Airyscan <https://www.lfd.uci.edu/globals/>`_,
   Image correlation analysis for the ZEISS(tm) LSM 880 Airyscan detector.

"""

from __future__ import annotations

__version__ = '2026.3.3'

__all__ = [
    'API',
    'IcsError',
    '__version__',
    'bins2times',
    'circle',
    'ipcf_nlsp_1dpcf',
    'logbins',
    'nextpow2',
    'nlsp',
    'numpy_correlate',
    'numpy_correlate1d',
    'points2distances',
    'radial',
    'rfftnd',
    'xyt',
    'yxt',
    'yxt_apcf',
    'yxt_correct_bleaching',
    'yxt_dft',
    'yxt_imsd',
    'yxt_ipcf',
    'yxt_lstics',
    'yxt_subtract_immobile',
    'zyx_deconv',
]

import ctypes
import math
import os
import warnings

import numpy


def API(dllname: str | None = None) -> ctypes.CDLL:  # noqa: N802
    """Return ctypes interface to functions of ICSx64 DLL."""
    from ctypes import (
        POINTER,
        c_char_p,
        c_double,
        c_float,
        c_int,
        c_int32,
        c_int64,
        c_size_t,
        c_ssize_t,
    )

    c_ssize_t_p = POINTER(c_ssize_t)
    c_double_p = POINTER(c_double)
    handle_t = c_size_t

    if dllname is None:
        dllname = os.path.join(os.path.dirname(__file__), 'ICSx64.dll')

    api = ctypes.CDLL(dllname)

    version = c_char_p.in_dll(api, 'ICS_VERSION').value
    assert version is not None
    api.VERSION = version.decode('ascii')

    api.FALSE = 0
    api.TRUE = 1

    api.OK = 0
    api.ERROR = -1
    api.VALUE_ERROR = -2
    api.MEMORY_ERROR = -3
    api.NOTIMPLEMENTED_ERROR = -4

    api.VALUE_ERROR1 = -201
    api.VALUE_ERROR2 = -202
    api.VALUE_ERROR3 = -203
    api.VALUE_ERROR4 = -204
    api.VALUE_ERROR5 = -205
    api.VALUE_ERROR6 = -206
    api.VALUE_ERROR7 = -207
    api.VALUE_ERROR8 = -208
    api.VALUE_ERROR9 = -209

    api.MODE_DEFAULT = 0
    api.MODE_TIME = 1  # do not center correlation results in axis 0 (time)
    api.MODE_FCS = 2  # normalize correlation results according to FCS
    api.MODE_CC = 4  # allocate second buffer for cross correlation

    api.AXIS0 = 1  # do not center correlation results in axis 0
    api.AXIS1 = 8  # do not center correlation results in axis 1
    api.AXIS2 = 16  # do not center correlation results in axis 2

    api.MASK_DEFAULT = 0
    api.MASK_ANY = 0  # any one value must be True
    api.MASK_FIRST = 1  # first mask value must be True
    api.MASK_CENTER = 2  # center mask value must be True
    api.MASK_ALL = 4  # all mask values must be True
    api.MASK_CLEAR = 32  # clear output if not calculated

    api.RADIUS = 1
    api.DIAMETER = 2

    api.NLSP_ND = 1
    api.NLSP_1DPCF = 100

    api.ICS_DECONV_DEFAULT = 1
    api.ICS_DECONV_RICHARDSON_LUCY = 1
    api.ICS_DECONV_WIENER = 2
    api.ICS_DECONV_NOPAD = 256

    api.DTYPES = {'l': 'i', 'i': 'i', 'h': 'h', 'H': 'H', 'd': 'd', 'f': 'f'}

    def ndpointer(
        *,
        dtype=None,
        ndim=None,
        shape=None,
        flags=None,
        null=False,
    ):
        """Return numpy.ctypes.ndpointer type that also accepts None/NULL."""
        cls = numpy.ctypeslib.ndpointer(dtype, ndim, shape, flags)
        if not null:
            return cls
        from_param_ = cls.from_param

        def from_param(cls, param):
            del cls  # unused
            if param is None:
                return param
            return from_param_(param)

        cls.from_param = classmethod(from_param)  # type: ignore[assignment]
        return cls

    def outer(a, b, /, skip=()):
        return ((x, y) for x in a for y in b if (x, y) not in skip)

    # rfft#d_ functions
    for nd in (1, 2, 3):
        rfftnd = f'rfft{nd}d_'

        func = getattr(api, rfftnd + 'new')
        func.argtypes = [c_ssize_t] * nd + [c_int]
        func.restype = handle_t

        func = getattr(api, rfftnd + 'del')
        func.argtypes = [handle_t]
        func.restype = None

        func = getattr(api, rfftnd + 'mode')
        func.argtypes = [handle_t, c_int]
        func.restype = None

        for i, o in outer('dfihH', 'df'):
            ai = ndpointer(dtype=i, ndim=nd)
            ao = ndpointer(dtype=o, ndim=nd)

            func = getattr(api, rfftnd + f'autocorrelate_{i}{o}')
            func.argtypes = [handle_t, ai, ao, c_ssize_t_p]
            func.restype = c_int

            func = getattr(api, rfftnd + f'crosscorrelate_{i}{o}')
            func.argtypes = [handle_t, ai, ai, ao, c_ssize_t_p, c_ssize_t_p]
            func.restype = c_int

    # rfft2d_lstics
    for o in 'fd':
        func = getattr(api, f'rfft2d_lstics_{o}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype='d', ndim=2),  # data
            ndpointer(dtype='d', ndim=2, null=True),  # channel
            c_ssize_t_p,  # strides
            ndpointer(dtype=o, ndim=2),  # out
            c_ssize_t_p,  # outstrides
            ndpointer(dtype='intp', ndim=1),  # line
            ndpointer(dtype='intp', ndim=1),  # bins
            c_ssize_t,  # nbins
            c_double,  # filter
        ]

    # rfft3d_imsd
    for o in 'fd':
        func = getattr(api, f'rfft3d_imsd_{o}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype='d', ndim=3),  # data
            ndpointer(dtype='d', ndim=3, null=True),  # channel
            c_ssize_t_p,  # strides
            ndpointer(dtype=o, ndim=3),  # out
            c_ssize_t_p,  # outstrides
            ndpointer(dtype='intp', ndim=1),  # bins
            c_ssize_t,  # nbins
            c_double,  # filter
        ]

    # nlsp class
    # TODO: test nlsp_ functions
    api.nlsp_new.restype = handle_t
    api.nlsp_new.argtypes = [c_int, c_ssize_t_p]

    api.nlsp_del.restype = None
    api.nlsp_del.argtypes = [handle_t]

    api.nlsp_get.restype = c_int
    api.nlsp_get.argtypes = [
        handle_t,
        c_ssize_t_p,
        c_ssize_t_p,
        c_double_p,
        c_double_p,
    ]

    api.nlsp_set.restype = c_int
    api.nlsp_set.argtypes = [
        handle_t,
        c_ssize_t,
        c_ssize_t,
        ndpointer(dtype='float64', shape=(6,), null=True),
        c_double,
        c_double,
    ]

    for dt in 'fd':
        func = getattr(api, f'nlsp_eval_{dt}')
        func.restype = c_int
        func.argtypes = [handle_t, ndpointer(dtype=dt), c_ssize_t_p]

        func = getattr(api, f'nlsp_solve_{dt}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype=dt),  # data
            c_ssize_t_p,  # strides
            ndpointer(dtype='float64', ndim=1),  # extra
            ndpointer(dtype='float64', ndim=1, null=True),  # guess
            ndpointer(dtype='float64', ndim=1, null=True),  # bounds
            ndpointer(dtype='float64', ndim=1, null=True),  # datasolution
        ]

    # yxt class
    api.yxt_new.restype = handle_t
    api.yxt_new.argtypes = [c_ssize_t_p]

    api.yxt_del.restype = None
    api.yxt_del.argtypes = [handle_t]

    api.yxt_get_buffer.restype = c_double_p
    api.yxt_get_buffer.argtypes = [handle_t, c_ssize_t_p, c_ssize_t_p]

    for ti, to in outer('dfihH', 'df'):
        # yxt_ipcf_*
        func = getattr(api, f'yxt_ipcf_{ti}{to}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype=ti, ndim=3),  # data
            ndpointer(dtype=ti, ndim=3, null=True),  # channel
            c_ssize_t_p,  # strides
            ndpointer(dtype=to, ndim=4),  # out
            c_ssize_t_p,  # outstrides
            ndpointer(dtype='intp', ndim=1),  # points
            c_ssize_t,  # npoints
            ndpointer(dtype='intp', ndim=1),  # bins
            c_ssize_t,  # nbins
            c_double,  # threshold
            c_double,  # filter
            c_int,  # nthreads
        ]

        # yxt_apcf_*
        func = getattr(api, f'yxt_apcf_{ti}{to}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype=ti, ndim=2),  # data
            c_ssize_t_p,  # strides
            ndpointer(dtype=to, ndim=3),  # out
            c_ssize_t_p,  # outstrides
            ndpointer(dtype='intp', ndim=1),  # bins
            c_ssize_t,  # nbins
            c_int,  # autocorr
            c_double,  # filter
            c_int,  # nthreads
        ]

        # yxt_imsd_*
        func = getattr(api, f'yxt_imsd_{ti}{to}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype=ti, ndim=3, null=True),  # data
            c_ssize_t_p,  # strides
            ndpointer(dtype=ti, ndim=3, null=True),  # data1
            c_ssize_t_p,  # strides1
            ndpointer(dtype='int32', ndim=2, null=True),  # mask
            c_ssize_t_p,  # maskstrides
            c_int,  # maskmode
            ndpointer(dtype=to, ndim=5),  # out
            c_ssize_t_p,  # outstrides
            c_ssize_t_p,  # blocks
            ndpointer(dtype='intp', ndim=1, null=True),  # bins
            c_ssize_t,  # nbins
            c_double,  # filter
            c_int,  # nthreads
        ]

        # yxt_lstics_*
        func = getattr(api, f'yxt_lstics_{ti}{to}')
        func.restype = c_int
        func.argtypes = [
            handle_t,
            ndpointer(dtype=ti, ndim=3, null=True),  # data
            c_ssize_t_p,  # strides
            ndpointer(dtype=ti, ndim=3, null=True),  # data1
            c_ssize_t_p,  # strides1
            ndpointer(dtype='int32', ndim=2, null=True),  # mask
            c_ssize_t_p,  # maskstrides
            c_int,  # maskmode
            ndpointer(dtype=to, ndim=5),  # out
            c_ssize_t_p,  # outstrides
            ndpointer(dtype='intp', ndim=3),  # lines
            c_ssize_t_p,  # lineshape
            c_ssize_t_p,  # blocks
            ndpointer(dtype='intp', ndim=1),  # bins
            c_ssize_t,  # nbins
            c_double,  # filter
            c_int,  # nthreads
        ]

    # ipcf_nlsp_1dpcf
    for dt in 'f':
        func = getattr(api, f'ipcf_nlsp_1dpcf_{dt}')
        func.restype = c_int
        func.argtypes = [
            ndpointer(dtype=dt, ndim=4),  # ipcf
            c_ssize_t_p,  # shape
            c_ssize_t_p,  # strides
            ndpointer(dtype=dt, ndim=1),  # times
            ndpointer(dtype=dt, ndim=1),  # distances
            ndpointer(dtype=dt, ndim=1),  # args
            ndpointer(dtype=dt, ndim=1, null=True),  # bounds
            ndpointer(dtype=dt, ndim=4, null=True),  # ix
            c_ssize_t_p,  # stridesx
            ndpointer(dtype=dt, ndim=4, null=True),  # ifx
            c_ssize_t_p,  # stridesfx
            ndpointer(dtype=dt, ndim=4, null=True),  # status
            c_ssize_t_p,  # stridestatus
            ndpointer(dtype=dt, ndim=1, null=True),  # settings
            c_int,  # average (bool)
            c_int,  # nthreads
        ]

    # subtract_immobile, yxt_correct_bleaching
    for dt in 'ihH':
        ai = ndpointer(dtype=dt, ndim=3)

        func = getattr(api, f'yxt_subtract_immobile_{dt}')
        func.argtypes = [ai, c_ssize_t_p, c_ssize_t_p, c_int]
        func.restype = c_int

        func = getattr(api, f'yxt_correct_bleaching_{dt}')
        func.argtypes = [
            ai,
            c_ssize_t_p,
            c_ssize_t_p,
            ndpointer(dtype='double', ndim=2),
            c_ssize_t_p,
            c_double,
            c_int,
        ]
        func.restype = c_int

    # yxt_dft
    for ti, to in outer('dfihH', 'df', skip={('d', 'f'), ('f', 'd')}):
        func = getattr(api, f'yxt_dft_{ti}{to}')
        func.restype = c_int
        func.argtypes = [
            ndpointer(dtype=ti, ndim=3),  # data
            c_ssize_t_p,  # shape
            c_ssize_t_p,  # strides
            ndpointer(dtype=to, ndim=3),  # out
            c_ssize_t_p,  # outshape
            c_ssize_t_p,  # outstrides
            c_int,  # nthreads
        ]

    # zyx_deconv
    try:
        for ti, to in (('f', 'f'), ('d', 'd'), ('H', 'f')):  # outer('fH', 'f')
            func = getattr(api, f'zyx_deconv_{ti}{to}')
            func.restype = c_int
            func.argtypes = [
                ndpointer(dtype=ti, ndim=3),  # image
                c_ssize_t_p,  # shape
                c_ssize_t_p,  # strides
                ndpointer(dtype=ti, ndim=3),  # psf
                c_ssize_t_p,  # shape
                c_ssize_t_p,  # strides
                ndpointer(dtype=to, ndim=3),  # out
                c_ssize_t_p,  # outshape
                c_ssize_t_p,  # outstrides
                c_int,  # niter
                c_int,  # mode
                c_int,  # nthreads
            ]
    except AttributeError:
        pass

    # helper functions
    api.radial.restype = c_ssize_t
    api.radial.argtypes = [
        ndpointer(dtype='intp', ndim=3),
        c_ssize_t,
        c_ssize_t,
        ndpointer(dtype='float64', ndim=1),
        c_int,
    ]

    api.circle.restype = c_ssize_t
    api.circle.argtypes = [
        c_ssize_t,
        ndpointer(dtype='intp', ndim=2, null=True),
        c_ssize_t,
    ]

    api.logbins.restype = c_ssize_t
    api.logbins.argtypes = [
        c_ssize_t,
        c_ssize_t,
        ndpointer(dtype='intp', ndim=1),
    ]

    api.points2distances_f.restype = c_float
    api.points2distances_f.argtypes = [
        ndpointer(dtype='intp', ndim=2),
        c_ssize_t,
        c_float,
        ndpointer(dtype='float32', ndim=1),
    ]

    api.points2distances_d.restype = c_double
    api.points2distances_d.argtypes = [
        ndpointer(dtype='intp', ndim=2),
        c_ssize_t,
        c_double,
        ndpointer(dtype='float64', ndim=1),
    ]

    api.bins2times_f.restype = c_int
    api.bins2times_f.argtypes = [
        ndpointer(dtype='intp', ndim=1),
        c_ssize_t,
        c_float,
        ndpointer(dtype='float32', ndim=1),
    ]

    api.bins2times_d.restype = c_int
    api.bins2times_d.argtypes = [
        ndpointer(dtype='intp', ndim=1),
        c_ssize_t,
        c_double,
        ndpointer(dtype='float64', ndim=1),
    ]

    api.nextpow2_i.restype = c_int32
    api.nextpow2_i.argtypes = [c_int32]

    api.nextpow2_q.restype = c_int64
    api.nextpow2_q.argtypes = [c_int64]

    return api


API = API()  # type: ignore[assignment]


class IcsError(RuntimeError):
    """Exception raised when ICSx64 DLL functions returns error code."""

    def __init__(self, func, err, /):
        msg = {
            None: 'NULL',
            API.OK: 'OK',
            API.ERROR: 'ERROR',
            API.VALUE_ERROR: 'VALUE_ERROR',
            API.MEMORY_ERROR: 'MEMORY_ERROR',
            API.NOTIMPLEMENTED_ERROR: 'NOTIMPLEMENTED_ERROR',
            API.VALUE_ERROR1: 'VALUE_ERROR1',
            API.VALUE_ERROR2: 'VALUE_ERROR2',
            API.VALUE_ERROR3: 'VALUE_ERROR3',
            API.VALUE_ERROR4: 'VALUE_ERROR4',
            API.VALUE_ERROR5: 'VALUE_ERROR5',
            API.VALUE_ERROR6: 'VALUE_ERROR6',
            API.VALUE_ERROR7: 'VALUE_ERROR7',
            API.VALUE_ERROR8: 'VALUE_ERROR8',
            API.VALUE_ERROR9: 'VALUE_ERROR9',
        }.get(err, f'unknown error {err}')
        RuntimeError.__init__(self, f'{func.__name__} returned {msg}')


class rfftnd:
    """Wrapper for rfft1d_, rfft2d_, or rfft3d_ correlation functions."""

    def __init__(self, shape, mode):
        """Initialize rfft correlation handle.

        Parameters
        ----------
        shape : sequence of int
            Dimensions of data array (length 1, 2, or 3).
        mode : int
            Combination of API.MODE_* flags controlling correlation behavior.

        """
        self._handle = None
        self._rfftnd = f'rfft{len(shape)}d_'
        func = getattr(API, self._rfftnd + 'new')
        self._handle = func(*shape, mode)
        if self._handle == 0:
            raise IcsError(func, None)
        self._mode = mode

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        self._mode = value
        func = getattr(API, self._rfftnd + 'mode')
        func(self._handle, value)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self._handle:
            func = getattr(API, self._rfftnd + 'del')
            func(self._handle)
        self._handle = None

    def autocorrelate(self, a, /, *, out=None):
        """Compute in-place autocorrelation of array a.

        The result overwrites `a` when `out` is None.

        """
        if out is None:
            out = a
        elif not out.flags['C_CONTIGUOUS']:
            raise ValueError('out must be contiguous')
        func = getattr(
            API,
            f'{ self._rfftnd}autocorrelate_'
            f'{API.DTYPES[a.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )
        status = func(self._handle, a, out, a.ctypes.strides)
        if status:
            raise IcsError(func, status)

    def crosscorrelate(self, a, b, /, *, out=None):
        """Compute cross-correlation of arrays a and b.

        The result overwrites `a` when `out` is None.

        """
        if out is None:
            out = a
        elif not out.flags['C_CONTIGUOUS']:
            raise ValueError('out must be contiguous')
        func = getattr(
            API,
            f'{self._rfftnd}crosscorrelate_'
            f'{API.DTYPES[a.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )
        status = func(
            self._handle, a, b, out, a.ctypes.strides, b.ctypes.strides
        )
        if status:
            raise IcsError(func, status)

    def lstics(
        self,
        data,
        /,
        line,
        bins,
        *,
        channel=None,
        out=None,
        smooth=0.0,
    ):
        """Compute line spatio-temporal image correlation (rfft2d only).

        Parameters
        ----------
        data : ndarray, shape (y, x)
            Input image data as float64.
        line : ndarray of intp
            Pixel coordinates of scan line.
        bins : ndarray of intp
            Logarithmically spaced time-lag bin indices.
        channel : ndarray or None
            Optional second channel with same shape as data.
        out : ndarray
            Output array; required.
        smooth : float
            Gaussian smoothing sigma applied before correlation.

        """
        if self._rfftnd != 'rfft2d_':
            raise ValueError('lstics requires rfft2d')
        if out is None:
            raise ValueError('out is required for lstics')
        func = getattr(
            API,
            f'rfft2d_lstics_{API.DTYPES[out.dtype.char]}',
        )
        status = func(
            self._handle,
            data,
            channel,
            data.ctypes.strides,
            out,
            out.ctypes.strides,
            line,
            bins,
            len(bins),
            smooth,
        )
        if status:
            raise IcsError(func, status)

    def imsd(
        self,
        data,
        /,
        bins,
        *,
        channel=None,
        out=None,
        smooth=0.0,
    ):
        """Compute image mean square displacement (rfft3d only).

        Parameters
        ----------
        data : ndarray, shape (y, x, t)
            Input image time-series as float64.
        bins : ndarray of intp
            Logarithmically spaced time-lag bin indices.
        channel : ndarray or None
            Optional second channel with same shape as data.
        out : ndarray
            Output array; required.
        smooth : float
            Gaussian smoothing sigma applied before correlation.

        """
        if self._rfftnd != 'rfft3d_':
            raise ValueError('imsd requires rfft3d')
        if out is None:
            raise ValueError('out is required for imsd')
        func = getattr(
            API,
            f'rfft3d_imsd_{API.DTYPES[out.dtype.char]}',
        )
        status = func(
            self._handle,
            data,
            channel,
            data.ctypes.strides,
            out,
            out.ctypes.strides,
            bins,
            len(bins),
            smooth,
        )
        if status:
            raise IcsError(func, status)


class nlsp:
    """Wrapper class for nlsp_ functions.

    Solver of non-linear least squares problem with linear boundary
    constraints using RCI and the Trust-Region algorithm.

    Only the "1D pair correlation function" diffusion model is currently
    supported.

    """

    def __init__(self, shape, /, *, model='1dpcf'):
        self._handle = None
        warnings.warn('the nlsp class is untested', stacklevel=2)
        model = {API.NLSP_1DPCF: API.NLSP_1DPCF, '1dpcf': API.NLSP_1DPCF}[
            model
        ]
        shape = (ctypes.c_ssize_t * len(shape))(*shape)
        self._handle = API.nlsp_new(model, shape)
        if self._handle == 0:
            raise IcsError(API.nlsp_new, None)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self._handle:
            API.nlsp_del(self._handle)
        self._handle = None

    def solve(
        self,
        data,
        /,
        extra,
        *,
        guess=None,
        bounds=None,
        solution=None,
    ):
        """Solve nonlinear least squares problem.

        For the 1dpcf model, the 'extra' argument contains the xaxis values,
        the w2 parameter, and the squared distance.

        """
        func = getattr(API, f'nlsp_solve_{API.DTYPES[data.dtype.char]}')
        status = func(
            self._handle,
            data,
            data.ctypes.strides,
            extra,
            guess,
            bounds,
            solution,
        )
        if status:
            raise IcsError(func, status)

    def eval(self, data, /):
        """Evaluate function using current solution vector."""
        func = getattr(API, f'nlsp_eval_{API.DTYPES[data.dtype.char]}')
        status = func(self._handle, data, data.ctypes.strides)
        if status:
            raise IcsError(func, status)

    def get(self):
        """Return solution statuses.

        Return number of iterations, stop criterion, initial residual, and
        final residual.

        """
        it = ctypes.c_ssize_t()
        st_cr = ctypes.c_ssize_t()
        r1 = ctypes.c_double()
        r2 = ctypes.c_double()
        status = API.nlsp_get(self._handle, it, st_cr, r1, r2)
        if status:
            raise IcsError(API.nlsp_get, status)
        return it.value, st_cr.value, r1.value, r2.value

    def set(self, *, iter1=0, iter2=0, eps=None, eps_jac=0.0, rs=0.0):
        """Set solver parameters.

        Parameters
        ----------
        iter1 : int
            Maximum number of iterations for the trial step calculation.
        iter2 : int
            Maximum number of iterations of the Trust-Region algorithm.
        eps : ndarray of float64, shape (6,) or None
            Stopping criteria vector. None uses DLL defaults.
        eps_jac : float
            Relative step size used for Jacobian approximation.
        rs : float
            Initial step bound. Zero uses DLL default.

        """
        status = API.nlsp_set(self._handle, iter1, iter2, eps, eps_jac, rs)
        if status:
            raise IcsError(API.nlsp_set, status)


class yxt:
    """Wrapper class for yxt_ image correlation functions.

    Data arrays are expected in (y, x, t) axis order, where y and x are
    spatial dimensions and t is the time (frame) dimension.

    """

    def __init__(self, shape, /):
        """Initialize yxt correlation handle.

        Parameters
        ----------
        shape : sequence of int
            Shape (y, x, t) of image time-series to be analyzed.
            Time dimension t is truncated to nearest power of two.

        """
        self._handle = None
        shape = (ctypes.c_ssize_t * len(shape))(*shape)
        self._handle = API.yxt_new(shape)
        if self._handle == 0:
            raise IcsError(API.yxt_new, None)
        # retrieve truncated shape
        API.yxt_get_buffer(self._handle, shape, None)
        self.shape = int(shape[0]), int(shape[1]), int(shape[2])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self._handle:
            API.yxt_del(self._handle)
        self._handle = None

    def ipcf(
        self,
        data,
        /,
        points,
        bins,
        *,
        channel=None,
        out=None,
        threshold=0.0,
        smooth=0.0,
        nthreads=0,
        verbose=False,
    ):
        """Compute image pair correlation function.

        Parameters
        ----------
        data : ndarray, shape (y, x, t)
            Input image time-series.
        points : ndarray of intp, shape (npoints, 2)
            (x, y) pixel offsets from circle or custom geometry.
        bins : ndarray of intp
            Logarithmically spaced time-lag bin indices.
        channel : ndarray or None
            Optional cross-correlation channel with same shape and strides
            as data.
        out : ndarray or None
            Output array of shape (y', x', npoints, nbins). Allocated if None.
        threshold : float
            Pixels below this intensity are excluded.
        smooth : float
            Gaussian smoothing sigma applied before correlation.
        nthreads : int
            Number of OpenMP threads (0 = use all available).
        verbose : bool
            Print shape and stride information.

        Returns
        -------
        out : ndarray, shape (y', x', npoints, nbins)

        """
        if (
            data.ndim != 3
            or data.shape[0] != self.shape[0]
            or data.shape[1] != self.shape[1]
        ):
            raise ValueError('invalid data shape')

        if channel is not None and (
            channel.strides != data.strides or channel.shape != data.shape
        ):
            raise ValueError('invalid channel shape')

        npoints = points.shape[0]
        nbins = len(bins)

        x0, y0 = points.min(axis=0)
        x1, y1 = points.max(axis=0)
        outshape = (
            int(data.shape[0] - y1 + y0),
            int(data.shape[1] - x1 + x0),
            npoints,
            nbins,
        )
        if out is None:
            out = numpy.zeros(shape=outshape, dtype='float32')
        elif out.ndim != 4 or out.size < product(outshape):
            raise ValueError

        func = getattr(
            API,
            'yxt_ipcf_'
            f'{API.DTYPES[data.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )

        if verbose:
            print('data shape =', data.shape)
            print('data strides =', data.strides)
            print('out shape =', out.shape)
            print('out strides =', out.strides)

        status = func(
            self._handle,
            data,
            channel,
            data.ctypes.strides,
            out,
            out.ctypes.strides,
            points.flatten(),
            npoints,
            bins,
            nbins,
            threshold,
            smooth,
            nthreads,
        )
        if status != 0:
            raise IcsError(func, status)

        return out

    def apcf(
        self,
        data,
        /,
        bins,
        *,
        out=None,
        autocorr=True,
        smooth=0.0,
        nthreads=0,
        verbose=False,
    ):
        """Compute Airy detector pair correlation function.

        Requires yxt handle to have been created with shape (1, ndet, t).

        Parameters
        ----------
        data : ndarray, shape (ndet, t)
            Detector element time traces.
        bins : ndarray of intp
            Logarithmically spaced time-lag bin indices.
        out : ndarray or None
            Output array of shape (ndet, ndet[, -1], nbins). Allocated if None.
        autocorr : bool
            If True, include auto-correlation on diagonal.
        smooth : float
            Gaussian smoothing sigma applied before correlation.
        nthreads : int
            Number of OpenMP threads (0 = use all available).
        verbose : bool
            Print shape and stride information.

        Returns
        -------
        out : ndarray, shape (ndet, ndet or ndet-1, nbins)

        """
        if (
            data.ndim != 2
            or data.shape[0] != self.shape[1]
            or self.shape[0] != 1
        ):
            raise ValueError('invalid data shape')

        nbins = len(bins)
        if autocorr:
            outshape = data.shape[0], data.shape[0], nbins
        else:
            outshape = data.shape[0], data.shape[0] - 1, nbins
        if out is None:
            out = numpy.zeros(shape=outshape, dtype='float32')
        elif out.ndim != 3 or out.size < product(outshape):
            raise ValueError

        func = getattr(
            API,
            'yxt_apcf_'
            f'{API.DTYPES[data.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )

        if verbose:
            print('data shape =', data.shape)
            print('data strides =', data.strides)
            print('out shape =', out.shape)
            print('out strides =', out.strides)

        status = func(
            self._handle,
            data,
            data.ctypes.strides,
            out,
            out.ctypes.strides,
            bins,
            nbins,
            autocorr,
            smooth,
            nthreads,
        )
        if status != 0:
            raise IcsError(func, status)

        return out

    def imsd(
        self,
        data,
        /,
        block,
        bins,
        *,
        channel=None,
        out=None,
        mask=None,
        mask_mode=None,
        smooth=0.0,
        nthreads=0,
        verbose=False,
    ):
        """Compute image mean square displacement.

        Parameters
        ----------
        data : ndarray, shape (y, x, t)
            Input image time-series.
        block : sequence of 4 int
            Block shape and step (ysize, xsize, ystep, xstep).
        bins : int or ndarray of intp
            Number of logarithmic time-lag bins, or explicit bin indices.
        channel : ndarray or None
            Optional cross-correlation channel.
        out : ndarray or None
            Output array of shape (ny, nx, ysize, xsize, nbins). Allocated
            if None.
        mask : ndarray of int32 or None
            Optional spatial mask of shape (y, x).
        mask_mode : int or None
            Combination of API.MASK_* flags. Defaults to MASK_ANY|MASK_CLEAR.
        smooth : float
            Gaussian smoothing sigma applied before correlation.
        nthreads : int
            Number of OpenMP threads (0 = use all available).
        verbose : bool
            Print shape and stride information.

        Returns
        -------
        out : ndarray, shape (ny, nx, ysize, xsize, nbins)

        """
        if (
            data.ndim != 3
            or data.shape[0] != self.shape[0]
            or data.shape[1] != self.shape[1]
        ):
            raise ValueError('invalid data shape')

        if channel is not None:
            if channel.shape != data.shape:
                raise ValueError('invalid channel shape')
            channel_strides = channel.ctypes.strides
        else:
            channel_strides = None

        if mask is not None:
            if mask.shape[:2] != data.shape[:2]:
                raise ValueError('invalid mask shape')
            mask_strides = mask.ctypes.strides
        else:
            mask_strides = None

        if mask_mode is None:
            mask_mode = API.MASK_ANY | API.MASK_CLEAR

        if len(block) != 4:
            raise ValueError

        try:
            nbins = int(bins)
            bins = None
        except Exception:
            nbins = len(bins)

        outshape = (
            int((data.shape[0] - block[0]) // block[2] + 1),
            int((data.shape[1] - block[1]) // block[3] + 1),
            int(block[0]),
            int(block[1]),
            nbins,
        )
        if out is None:
            out = numpy.zeros(shape=outshape, dtype='float32')
        elif out.ndim != 5 or out.size < product(outshape):
            raise ValueError('invalid out shape')

        block = (ctypes.c_ssize_t * 4)(*block)

        func = getattr(
            API,
            'yxt_imsd_'
            f'{API.DTYPES[data.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )

        if verbose:
            print('data shape =', data.shape)
            print('data strides =', data.strides)
            print('out shape =', out.shape)
            print('out strides =', out.strides)

        status = func(
            self._handle,
            data,
            data.ctypes.strides,
            channel,
            channel_strides,
            mask,
            mask_strides,
            mask_mode,
            out,
            out.ctypes.strides,
            block,
            bins,
            nbins,
            smooth,
            nthreads,
        )
        if status != 0:
            raise IcsError(func, status)

        return out

    def lstics(
        self,
        data,
        /,
        block,
        lines,
        bins,
        *,
        channel=None,
        out=None,
        mask=None,
        mask_mode=None,
        smooth=0.0,
        nthreads=0,
        verbose=False,
    ):
        """Compute line spatio-temporal image correlation spectroscopy.

        Parameters
        ----------
        data : ndarray, shape (y, x, t)
            Input image time-series.
        block : sequence of 4 int
            Block shape and step (ysize, xsize, ystep, xstep).
        lines : ndarray of intp, shape (nlines, linelength, 2)
            Pixel coordinates of scan lines (from `radial`).
        bins : ndarray of intp
            Logarithmically spaced time-lag bin indices.
        channel : ndarray or None
            Optional cross-correlation channel.
        out : ndarray or None
            Output array of shape (ny, nx, nlines, linelength, nbins).
            Allocated if None.
        mask : ndarray of int32 or None
            Optional spatial mask of shape (y, x).
        mask_mode : int or None
            Combination of API.MASK_* flags. Defaults to MASK_ANY|MASK_CLEAR.
        smooth : float
            Gaussian smoothing sigma applied before correlation.
        nthreads : int
            Number of OpenMP threads (0 = use all available).
        verbose : bool
            Print shape and stride information.

        Returns
        -------
        out : ndarray, shape (ny, nx, nlines, linelength, nbins)

        """
        if (
            data.ndim != 3
            or data.shape[0] != self.shape[0]
            or data.shape[1] != self.shape[1]
        ):
            raise ValueError('invalid data shape')

        if channel is not None:
            if channel.shape != data.shape:
                raise ValueError('invalid channel shape')
            channel_strides = channel.ctypes.strides
        else:
            channel_strides = None

        if mask is not None:
            if mask.shape[:2] != data.shape[:2]:
                raise ValueError('invalid mask shape')
            mask_strides = mask.ctypes.strides
        else:
            mask_strides = None

        if mask_mode is None:
            mask_mode = API.MASK_ANY | API.MASK_CLEAR

        if len(block) != 4:
            raise ValueError

        nbins = len(bins)

        outshape = (
            int((data.shape[0] - block[0]) // block[2] + 1),
            int((data.shape[1] - block[1]) // block[3] + 1),
            int(lines.shape[0]),
            int(lines.shape[1]),
            int(nbins),
        )
        if out is None:
            out = numpy.zeros(shape=outshape, dtype='float32')
        elif out.ndim != 5 or out.size < product(outshape):
            raise ValueError('invalid out shape')

        block = (ctypes.c_ssize_t * 4)(*block)
        lines_shape = (ctypes.c_ssize_t * 3)(*lines.shape)

        func = getattr(
            API,
            'yxt_lstics_'
            f'{API.DTYPES[data.dtype.char]}{API.DTYPES[out.dtype.char]}',
        )

        if verbose:
            print('data shape =', data.shape)
            print('data strides =', data.strides)
            print('out shape =', out.shape)
            print('out strides =', out.strides)
            print('lines shape =', lines.shape)

        status = func(
            self._handle,
            data,
            data.ctypes.strides,
            channel,
            channel_strides,
            mask,
            mask_strides,
            mask_mode,
            out,
            out.ctypes.strides,
            lines,
            lines_shape,
            block,
            bins,
            nbins,
            smooth,
            nthreads,
        )
        if status != 0:
            raise IcsError(func, status)

        return out


xyt = yxt  # backward compatibility


def yxt_ipcf(
    data,
    /,
    *,
    radius=4,
    nbins=32,
    smooth=0.7,
    threshold=0.0,
    nthreads=0,
):
    """Simplified image pair correlation function."""
    # make time axis last dimension
    # if data.shape[0] > 8 * data.shape[2]:
    #     data = numpy.moveaxis(data, 0, -1)
    _height, _width, ntimes = data.shape

    # truncate time axis to power of two
    ntimes = 2 ** int(math.log2(ntimes))
    data = data[..., :ntimes]

    bins = logbins(ntimes // 2, nbins)
    # nbins = bins.shape[0]

    points = circle(radius)
    # npoints = points.shape[0]

    with yxt(data.shape) as handle:
        result = handle.ipcf(
            data,
            points,
            bins,
            smooth=smooth,
            threshold=threshold,
            nthreads=nthreads,
        )
    return result, bins, points


def yxt_apcf(
    data,
    /,
    *,
    nbins=256,
    autocorr=True,
    smooth=0.7,
    nthreads=0,
):
    """Simplified airy detector pair correlation."""
    width, ntimes = data.shape

    # truncate time axis to power of two
    ntimes = 2 ** int(math.log2(ntimes))
    data = data[..., :ntimes]

    bins = logbins(ntimes // 2, nbins)
    # nbins = bins.shape[0]

    with yxt((1, width, ntimes)) as handle:
        result = handle.apcf(
            data, bins, autocorr=autocorr, smooth=smooth, nthreads=nthreads
        )
    return result, bins


def yxt_imsd(
    data,
    /,
    *,
    block=(32, 32, 4, 4),
    bins=16,
    smooth=0.0,
    nthreads=0,
):
    """Simplified image mean square displacement."""
    with yxt(data.shape) as handle:
        return handle.imsd(data, block, bins, smooth=smooth, nthreads=nthreads)


def yxt_lstics(
    data,
    /,
    *,
    block=(16, 16, 1, 1),
    nlines=16,
    linelength=8,
    nbins=16,
    smooth=0.0,
    nthreads=0,
):
    """Simplified line spatio temporal image correlation spectroscopy."""
    _height, _width, ntimes = data.shape

    # truncate time axis to power of two
    ntimes = 2 ** int(math.log2(ntimes))
    data = data[..., :ntimes]

    bins = logbins(ntimes // 2, nbins)
    # nbins = bins.shape[0]

    lines = radial(nlines, linelength)

    with yxt(data.shape) as handle:
        result = handle.lstics(
            data, block, lines, bins, smooth=smooth, nthreads=nthreads
        )
    return result, bins, lines


def yxt_subtract_immobile(a, /, *, nthreads=0):
    """Subtract immobile fraction from (y, x, t) image time-series.

    Modifies `a` in place by removing the time-averaged (immobile) component
    from each pixel.

    """
    if a.ndim != 3:
        raise ValueError('input must be three dimensional')
    func = getattr(API, f'yxt_subtract_immobile_{API.DTYPES[a.dtype.char]}')
    status = func(a, a.ctypes.shape, a.ctypes.strides, nthreads)
    if status:
        raise IcsError(func, status)


def yxt_correct_bleaching(
    a,
    /,
    smooth,
    *,
    nthreads=0,
):
    """Correct photo-bleaching in (y, x, t) image time-series.

    Normalizes each pixel's time trace to remove the slow exponential decay
    due to photo-bleaching. Modifies `a` in place.

    Parameters
    ----------
    a : ndarray, shape (y, x, t)
        Image time-series (integer dtype: int16, uint16, or int32).
    smooth : float
        Gaussian smoothing sigma applied to bleaching curve before
        correction.
    nthreads : int
        Number of OpenMP threads (0 = use all available).

    Returns
    -------
    out : ndarray of float64, shape (y, x)
        Estimated bleaching correction factor per pixel.

    """
    if a.ndim != 3:
        raise ValueError('input must be three dimensional')
    func = getattr(API, f'yxt_correct_bleaching_{API.DTYPES[a.dtype.char]}')
    out = numpy.empty(a.shape[:2], 'float64')
    status = func(
        a,
        a.ctypes.shape,
        a.ctypes.strides,
        out,
        out.ctypes.strides,
        smooth,
        nthreads,
    )
    if status:
        raise IcsError(func, status)
    return out


def yxt_dft(
    data,
    /,
    *,
    samples=5,
    nthreads=0,
    asimages=True,
):
    """Compute 1D DFTs at fixed set of frequencies for each pixel.

    Evaluates the Goertzel DFT at `samples` logarithmically spaced frequencies
    along the time axis of a (y, x, t) image time-series.

    Parameters
    ----------
    data : ndarray, shape (y, x, t)
        Input image time-series.
    samples : int
        Number of frequency components to compute.
    nthreads : int
        Number of OpenMP threads (0 = use all available).
    asimages : bool
        If True (default), return shape (samples, y, x) so each frequency
            is separate image plane. If False, return shape (y, x, samples).
    -------
    result : ndarray, shape (samples, y, x) or (y, x, samples)

    """
    if data.ndim != 3:
        raise ValueError('input must be three dimensional')

    shape = data.shape
    odtype = 'float64' if data.dtype.char in 'dl' else 'float32'
    if asimages:
        outshape = samples, shape[0], shape[1]
    else:
        outshape = shape[0], shape[1], samples
    result = numpy.zeros(outshape, dtype=odtype)
    out = numpy.moveaxis(result, 0, -1) if asimages else result

    func = getattr(
        API,
        f'yxt_dft_{API.DTYPES[data.dtype.char]}{API.DTYPES[numpy.dtype(odtype).char]}',
    )

    status = func(
        data,
        data.ctypes.shape,
        data.ctypes.strides,
        out,
        out.ctypes.shape,
        out.ctypes.strides,
        nthreads,
    )
    if status:
        raise IcsError(func, status)
    return result


def zyx_deconv(
    image,
    psf,
    /,
    *,
    iterations=10,
    dtype='f4',
    mode=0,
    nthreads=0,
):
    """Deconvolve 3D image with point spread function.

    Parameters
    ----------
    image : ndarray, shape (z, y, x)
        Input 3D image volume.
    psf : ndarray, shape (z, y, x)
        Point spread function; must have same dtype as `image`.
    iterations : int
        Number of deconvolution iterations.
    dtype : str or dtype
        Output dtype. Defaults to float32.
    mode : int
        Algorithm and options. Use API.ICS_DECONV_RICHARDSON_LUCY (default)
        or API.ICS_DECONV_WIENER. Combine with API.ICS_DECONV_NOPAD to skip
        zero-padding.
    nthreads : int
        Number of OpenMP threads (0 = use all available).

    Returns
    -------
    out : ndarray, shape (z, y, x)

    """
    if image.ndim != 3 or psf.ndim != 3:
        raise ValueError('input must be three dimensional')

    out = numpy.zeros(image.shape, dtype=dtype)

    func = getattr(
        API,
        'zyx_deconv_'
        f'{API.DTYPES[image.dtype.char]}{API.DTYPES[out.dtype.char]}',
    )

    status = func(
        image,
        image.ctypes.shape,
        image.ctypes.strides,
        psf,
        psf.ctypes.shape,
        psf.ctypes.strides,
        out,
        out.ctypes.shape,
        out.ctypes.strides,
        iterations,
        mode,
        nthreads,
    )
    if status:
        raise IcsError(func, status)
    return out


def ipcf_nlsp_1dpcf(
    ipcf,
    times,
    distances,
    args,
    *,
    bounds=None,
    ix=None,
    ifx=None,
    status=None,
    settings=None,
    average=False,
    nthreads=0,
):
    """Fit diffusion to results of ipcf analysis."""
    if ipcf.ndim != 4:
        raise ValueError

    dtype = ipcf.dtype

    func = getattr(API, f'ipcf_nlsp_1dpcf_{API.DTYPES[dtype.char]}')

    args = numpy.array(args, dtype=dtype)
    if bounds is not None:
        bounds = numpy.array(bounds, dtype=dtype)

    if settings is not None:
        settings = numpy.array(settings, dtype=dtype)

    npoints = 1 if average else ipcf.shape[2]
    ifxshape = ipcf.shape[0], ipcf.shape[1], npoints, ipcf.shape[3]
    ixshape = ipcf.shape[0], ipcf.shape[1], npoints, 2
    statusshape = ipcf.shape[0], ipcf.shape[1], npoints, 4

    if ix is None:
        ix = numpy.zeros(shape=ixshape, dtype=dtype)
    elif ix is False:
        ix = None
    elif ix.ndim != 4 or ix.size < product(ixshape):
        raise ValueError

    if ifx is None:
        ifx = numpy.zeros(shape=ifxshape, dtype=dtype)
    elif ifx is False:
        ifx = None
    elif ifx.ndim != len(ifxshape) or ifx.size < product(ifxshape):
        raise ValueError

    if status is None:
        status = numpy.zeros(shape=statusshape, dtype=dtype)
    elif status is False:
        status = None
    elif status.ndim != len(statusshape) or status.size < product(statusshape):
        raise ValueError

    distances = numpy.array(distances, dtype=dtype).reshape(-1)
    distances = distances * distances

    ret = func(
        ipcf,
        ipcf.ctypes.shape,
        ipcf.ctypes.strides,
        times,
        distances,
        args,
        bounds,
        ix,
        None if ix is None else ix.ctypes.strides,
        ifx,
        None if ifx is None else ifx.ctypes.strides,
        status,
        None if status is None else status.ctypes.strides,
        settings,
        average,
        nthreads,
    )
    if ret != 0:
        raise IcsError(func, ret)

    return ix, ifx, status


def points2distances(points, /, pixelsize, *, dtype='float32'):
    """Return distances from points on integer grid."""
    func = getattr(
        API, f'points2distances_{API.DTYPES[numpy.dtype(dtype).char]}'
    )
    distances = numpy.zeros(points.shape[0], dtype)
    ret = func(points, points.shape[0], pixelsize, distances)
    if ret == -1.0:
        raise IcsError(func, ret)
    return distances


def radial(nlines, length, /, *, mode='radius', offset=None):
    """Return integer coordinates of line segments through center of circle."""
    if nlines < 1 or length < 1:
        raise ValueError('invalid nbins or length')
    points = numpy.zeros((nlines, length, 2), dtype='intp')
    mode = {
        API.RADIUS: API.RADIUS,
        API.DIAMETER: API.DIAMETER,
        'radius': API.RADIUS,
        'diameter': API.DIAMETER,
    }[mode]
    if offset is None:
        offset = numpy.zeros(2, 'double')
    else:
        offset = numpy.array(offset).astype('double')
        if offset.size < 2:
            raise ValueError('invalid offset')
    status = API.radial(points, nlines, length, offset, mode)
    if status <= 0:
        raise IcsError(API.radial, status)
    return points


def circle(radius, /, npoints=0):
    """Return (x, y) integer coordinates of points on circle."""
    if npoints <= 0:
        npoints = API.circle(radius, None, 0)
    points = numpy.zeros((npoints, 2), 'intp')
    if npoints == 0:
        return points  # TODO: passing empty array to API causes random crash
    ret = API.circle(radius, points, npoints)
    if ret < 0:
        raise IcsError(API.circle, ret)
    return points


def logbins(size, nbins):
    """Return logarithmically spaced integer bin indices up to size."""
    if nbins < 1:
        raise ValueError('invalid nbins')
    bins = numpy.zeros(nbins, 'intp')
    nbins = API.logbins(size, nbins, bins)
    if nbins < 0:
        raise IcsError(API.logbins, nbins)
    if nbins != bins.size:
        bins = bins[:nbins].copy()
    return bins


def bins2times(bins, frametime, *, dtype='float32'):
    """Return times from bins."""
    func = getattr(API, f'bins2times_{API.DTYPES[numpy.dtype(dtype).char]}')
    times = numpy.zeros(bins.shape[0], dtype)
    status = func(bins, bins.shape[0], frametime, times)
    if status < 0:
        raise IcsError(func, status)
    return times


def nextpow2(n, /):
    """Return next power of 2."""
    return API.nextpow2_q(n)


def numpy_correlate(a, b, /, *, mode=0, axes=None):
    """Return cross-correlation using numpy.fft.fftn."""
    a = numpy.fft.fftn(a)
    b = numpy.fft.fftn(b)
    index = (0,) * a.ndim
    scale = a[index].real * b[index].real / a.size
    a *= b.conj()
    a = numpy.fft.ifftn(a).real
    if axes or axes is None:
        a = numpy.fft.fftshift(a, axes)
    if mode & API.MODE_FCS:
        a /= scale
        a -= 1.0
    return a


def numpy_correlate1d(a, b, /, *, mode=0, axes=None):
    """Return cross-correlation using numpy.fft.rfft."""
    n = a.size
    a = numpy.fft.rfft(a)
    b = numpy.fft.rfft(b)
    scale = a[0].real * b[0].real / n
    a *= b.conj()
    a = numpy.fft.irfft(a).real
    if axes or axes is None:
        a = numpy.fft.fftshift(a, axes)
    if mode & API.MODE_FCS:
        a /= scale
        a -= 1.0
    return a


def product(iterable, /):
    """Return product of sequence of numbers."""
    prod = 1
    for i in iterable:
        prod *= int(i)
    return prod


# mypy: allow-untyped-defs, allow-untyped-calls
# mypy: disable-error-code="attr-defined"



# test_icsdll.py

# Copyright (c) 2016-2026, Christoph Gohlke
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the image correlation spectroscopy library ICSx64.dll.

:Version: 2026.3.3

"""

import math
import os
import sys
import time
from contextlib import contextmanager

import numpy
import pytest
from numpy.testing import assert_allclose, assert_array_equal

VERSION = '2026.3.3'
APIVERSION = '2026.3.3'
HERE = os.path.dirname(__file__) + '/'
RNG = numpy.random.default_rng(42)


def test_versions():
    """Test versions match."""
    ver = icsdll.__version__
    assert ver == VERSION
    assert API.VERSION == APIVERSION
    assert ver in __doc__
    assert ver in icsdll.__doc__


def test_logbins():
    """Test logbins function."""
    # fmt: off
    assert_array_equal(
        logbins(32, 32),
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 16, 18, 20, 22, 25, 28, 32]
    )
    assert_array_equal(
        logbins(8192, 32),
        [1, 2, 3, 4, 5, 7, 10, 13, 18, 24, 32, 43, 58, 78, 104, 139, 187, 250,
         334, 447, 598, 800, 1070, 1432, 1915, 2561, 3425, 4580, 6125, 8192]
    )
    # fmt: on
    with pytest.raises(icsdll.IcsError):
        logbins(2, 2)


def test_circle():
    """Test circle function."""
    assert_array_equal(circle(0), numpy.empty((0, 2), dtype='int64'))
    assert_array_equal(circle(1), [[1, 0], [0, 1], [-1, 0], [0, -1]])
    # fmt: off
    assert_array_equal(
        circle(6),
        [[6, 0], [6, 1], [6, 2], [5, 3], [4, 4], [3, 5], [2, 6], [1, 6],
         [0, 6], [-1, 6], [-2, 6], [-3, 5], [-4, 4], [-5, 3], [-6, 2],
         [-6, 1], [-6, 0], [-6, -1], [-6, -2], [-5, -3], [-4, -4], [-3, -5],
         [-2, -6], [-1, -6], [0, -6], [1, -6], [2, -6], [3, -5], [4, -4],
         [5, -3], [6, -2], [6, -1]]
    )
    # fmt: on


def test_radial():
    """Test radial function."""
    # fmt: off
    assert_array_equal(
        radial(8, 4),
        numpy.array(
            [0, 0, 0, 1, 0, 2, 0, 3, 0, 0, 1, 1, 1, 1, 2, 2, 0, 0, 1, 0, 2,
             0, 3, 0, 0, 0, 1, -1, 1, -1, 2, -2, 0, 0, 0, -1, 0, -2, 0, -3,
             0, 0, -1, -1, -1, -1, -2, -2, 0, 0, -1, 0, -2, 0, -3, 0, 0, 0,
             -1, 1, -1, 1, -2, 2],
            dtype='intp',
        ).reshape((8, 4, 2))[:, :, ::-1],
    )
    # fmt: on


def test_nextpow2():
    """Test nextpow2 function."""
    assert nextpow2(1) == 1
    assert nextpow2(2) == 2
    assert nextpow2(3) == 4
    assert nextpow2(4) == 4
    assert nextpow2(5) == 8
    assert nextpow2(7) == 8
    assert nextpow2(8) == 8
    assert nextpow2(9) == 16
    assert nextpow2(1023) == 1024
    assert nextpow2(1024) == 1024
    assert nextpow2(1025) == 2048
    assert nextpow2(2**20) == 2**20
    assert nextpow2(2**20 + 1) == 2**21


def test_bins2times():
    """Test bins2times function."""
    bins = logbins(32, 16)
    frametime = 0.01

    times_f = bins2times(bins, frametime, dtype='float32')
    assert times_f.dtype == numpy.float32
    assert len(times_f) == len(bins)
    # first time is (bins[0] - 1) / 2 * frametime
    assert_allclose(times_f[0], (bins[0] - 1) / 2.0 * frametime, atol=1e-7)
    # times should be monotonically non-decreasing, and strictly after first
    assert numpy.all(numpy.diff(times_f[1:]) > 0)

    times_d = bins2times(bins, frametime, dtype='float64')
    assert times_d.dtype == numpy.float64
    assert len(times_d) == len(bins)
    assert_allclose(times_d[0], (bins[0] - 1) / 2.0 * frametime, atol=1e-15)
    assert numpy.all(numpy.diff(times_d[1:]) > 0)

    # float32 and float64 results should be close
    assert_allclose(times_f, times_d, rtol=1e-5)


def test_rfftnd_mode_setter():
    """Test rfftnd.mode setter changes correlation behavior."""
    shape = (256,)
    test = RNG.random(shape)

    # compute reference with MODE_FCS
    a = test.copy()
    with rfftnd(shape, mode=API.MODE_FCS) as fft:
        fft.autocorrelate(a, out=None)
    ref_fcs = a.copy()

    # compute reference with mode 0
    a = test.copy()
    with rfftnd(shape, mode=0) as fft:
        fft.autocorrelate(a, out=None)
    ref_default = a.copy()

    # start with mode 0, switch to MODE_FCS via setter
    a = test.copy()
    with rfftnd(shape, mode=0) as fft:
        assert fft.mode == 0
        fft.mode = API.MODE_FCS
        assert fft.mode == API.MODE_FCS
        fft.autocorrelate(a, out=None)
    assert numpy.allclose(a, ref_fcs)
    assert not numpy.allclose(a, ref_default)


def test_numpy_correlate1d():
    """Test numpy_correlate1d matches numpy_correlate for 1D input."""
    shape = (256,)
    a = RNG.random(shape)
    b = RNG.random(shape)

    # default mode
    r1 = numpy_correlate(a.copy(), b.copy(), mode=0)
    r2 = numpy_correlate1d(a.copy(), b.copy(), mode=0)
    assert_allclose(r1, r2, rtol=1e-5)

    # FCS mode
    r1 = numpy_correlate(a.copy(), b.copy(), mode=API.MODE_FCS)
    r2 = numpy_correlate1d(a.copy(), b.copy(), mode=API.MODE_FCS)
    assert_allclose(r1, r2, rtol=1e-5)

    # no shift
    r1 = numpy_correlate(a.copy(), b.copy(), mode=0, axes=False)
    r2 = numpy_correlate1d(a.copy(), b.copy(), mode=0, axes=False)
    assert_allclose(r1, r2, rtol=1e-5)

    # autocorrelation
    r1 = numpy_correlate(a.copy(), a.copy(), mode=API.MODE_FCS)
    r2 = numpy_correlate1d(a.copy(), a.copy(), mode=API.MODE_FCS)
    assert_allclose(r1, r2, rtol=1e-5)

    # verify against DLL rfft1d
    result = numpy.empty_like(a)
    mode = API.MODE_FCS
    ac = a.copy()
    bc = b.copy()
    with rfftnd(shape, mode=mode | API.MODE_CC) as fft:
        fft.crosscorrelate(ac, bc, out=result)
    r2 = numpy_correlate1d(a.copy(), b.copy(), mode=mode)
    assert_allclose(r2, result, rtol=1e-5)


def test_rfftnd_autocorrelate444():
    """Test rfft3d autocorrelate with Enrico's data."""
    test = numpy.zeros((4, 4, 4))
    for k in range(4):
        for j in range(4):
            for i in range(4):
                test[k, j, i] = k * 4 * 4 + j * 4 + i

    a = test.copy()
    t = numpy.zeros((17, 29, 51), dtype='float64')
    t -= 1.0
    t[3:11:2, 19:15:-1, 37:49:3] = a
    b = t[3:11:2, 19:15:-1, 37:49:3]  # a view

    # fmt: off
    test_out = numpy.array(
        [[[0.30964979, 0.31015369, 0.31166541, 0.31015369],
          [0.31771227, 0.31821618, 0.31972789, 0.31821618],
          [0.34189972, 0.34240363, 0.34391534, 0.34240363],
          [0.31771227, 0.31821618, 0.31972789, 0.31821618]],
         [[-0.07734946, -0.07684555, -0.07533384, -0.07684555],
          [-0.06928697, -0.06878307, -0.06727135, -0.06878307],
          [-0.04509952, -0.04459562, -0.0430839, -0.04459562],
          [-0.06928697, -0.06878307, -0.06727135, -0.06878307]],
         [[-0.20634921, -0.2058453, -0.20433359, -0.2058453],
          [-0.19828672, -0.19778282, -0.1962711, -0.19778282],
          [-0.17409927, -0.17359536, -0.17208365, -0.17359536],
          [-0.19828672, -0.19778282, -0.1962711, -0.19778282]],
         [[-0.07734946, -0.07684555, -0.07533384, -0.07684555],
          [-0.06928697, -0.06878307, -0.06727135, -0.06878307],
          [-0.04509952, -0.04459562, -0.0430839, -0.04459562],
          [-0.06928697, -0.06878307, -0.06727135, -0.06878307]]]
    )
    # fmt: on

    with rfftnd((4, 4, 4), mode=3) as fft:
        fft.autocorrelate(a, out=None)
        assert numpy.allclose(a, test_out)

        out = numpy.empty_like(b)
        fft.autocorrelate(b, out=out)
        assert numpy.allclose(out, test_out)

        a = test.copy()
        fft.crosscorrelate(a, a.copy(), out=None)
        assert numpy.allclose(a, test_out)

        out = numpy.empty_like(b)
        fft.crosscorrelate(b, b.copy(), out=out)
        assert numpy.allclose(out, test_out)


def test_rfftnd_autocorrelate1d():
    """Test rfft1d autocorrelate function."""
    shape = (256,)
    test = RNG.random(shape)
    result = numpy.empty_like(test)

    # default scaling and shift
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)


def test_rfftnd_autocorrelate2d():
    """Test rfft2d autocorrelate function."""
    shape = 256, 128
    test = RNG.random(shape)
    result = numpy.empty_like(test)

    # default scaling and shift
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=(1,))
    assert numpy.allclose(c, result)

    # do not shift last axis
    mode = API.AXIS1
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=(0,))
    assert numpy.allclose(c, result)

    # do not shift any axis
    mode = API.AXIS0 | API.AXIS1
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)


def test_rfftnd_autocorrelate3d():
    """Test rfft3d autocorrelate function."""
    shape = 64, 256, 128
    test = RNG.random(shape)
    result = numpy.empty_like(test)

    # default scaling and shift
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=(1, 2))
    assert numpy.allclose(c, result)

    # do not shift last axis
    mode = API.AXIS2
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=(0, 1))
    assert numpy.allclose(c, result)

    # do not shift any axis
    mode = API.AXIS0 | API.AXIS1 | API.AXIS2
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    c = numpy_correlate(a, a, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = 0
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, a, mode=mode)
    assert numpy.allclose(c, result)


def test_rfftnd_crosscorrelate1d():
    """Test rfft1d crosscorrelate function."""
    shape = (256,)
    test = RNG.random(shape)
    result = numpy.empty_like(test)
    b = RNG.random(shape)

    # default scaling and shift
    mode = 0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = 0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)


def test_rfftnd_crosscorrelate2d():
    """Test rfft2d crosscorrelate function."""
    shape = 256, 128
    test = RNG.random(shape)
    result = numpy.empty_like(test)
    b = RNG.random(shape)

    # default scaling and shift
    mode = 0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=(1,))
    assert numpy.allclose(c, result)

    # do not shift last axis
    mode = API.AXIS1 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=(0,))
    assert numpy.allclose(c, result)

    # do not shift any axis
    mode = API.AXIS0 | API.AXIS1 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = 0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)


def test_rfftnd_crosscorrelate3d():
    """Test rfft3d crosscorrelate function."""
    shape = 64, 256, 128
    test = RNG.random(shape)
    result = numpy.empty_like(test)
    b = RNG.random(shape)

    # default scaling and shift
    mode = 0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # fcs scaling
    mode = API.MODE_FCS | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # do not shift first axis
    mode = API.AXIS0 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=(1, 2))
    assert numpy.allclose(c, result)

    # do not shift last axis
    mode = API.AXIS2 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=(0, 1))
    assert numpy.allclose(c, result)

    # do not shift any axis
    mode = API.AXIS0 | API.AXIS1 | API.AXIS2 | API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=result)
    a = test.copy()
    c = numpy_correlate(a, b, mode=0, axes=False)
    assert numpy.allclose(c, result)

    # inplace
    mode = API.MODE_CC
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=None)
    result = a
    a = test.copy()
    c = numpy_correlate(a, b, mode=mode)
    assert numpy.allclose(c, result)

    # test cross- gives same result as auto-correlate
    # do not shift first axis
    mode = API.AXIS0 | API.MODE_FCS
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.autocorrelate(a, out=result)
    a = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, a, out=None)
    assert numpy.allclose(a, result)
    a = test.copy()
    b = test.copy()
    with rfftnd(shape, mode=mode) as fft:
        fft.crosscorrelate(a, b, out=None)
    assert numpy.allclose(a, result)


def test_rfftnd_huge3d():
    """Test rfft3d crosscorrelate with large data."""
    shape = 1024, 512, 1024
    a = RNG.random(shape)
    b = RNG.random(shape)
    with timer(f'rfftnd.crosscorrelate {shape}'):
        with rfftnd(shape, mode=API.MODE_CC) as fft:
            fft.crosscorrelate(a, b, out=a)


def test_rfftnd_shapes3d():
    """Test rfft3d crosscorrelate with different shapes."""
    shape = 32, 32, 2**16
    a = RNG.random(shape)
    with timer(f'rfftnd.autocorrelate {shape}'):
        with rfftnd(shape, mode=0) as fft:
            fft.autocorrelate(a, out=a)

    shape = 2**16, 32, 32
    a = RNG.random(shape)

    with timer(f'rfftnd.autocorrelate {shape}'):
        with rfftnd(shape, mode=0) as fft:
            fft.autocorrelate(a, out=a)


def test_rfftnd_threads():
    """Test rfft3d autocorrelate with threads."""
    from concurrent.futures import ThreadPoolExecutor
    from multiprocessing import cpu_count

    cpucount = cpu_count() // 2

    shape = 256, 256, 256
    a = RNG.random(shape)

    def function(a):
        result = numpy.empty_like(a)
        with rfftnd(a.shape, mode=0) as fft:
            fft.autocorrelate(a, out=result)
        return result

    c = numpy_correlate(a, a, mode=0)
    with timer('rfftnd.autocorrelate threads'):
        with ThreadPoolExecutor(cpucount) as executor:
            result = list(executor.map(function, [a] * cpucount))
    for r in result:
        assert numpy.allclose(c, r)


def test_rfftnd_simulation_channel():
    """Test rfft3d correlate with simulation data."""
    data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    data.shape = -1, 64, 64
    tsize = 2 ** int(math.log2(data.shape[0]))
    data = data[:tsize]
    # data = numpy.moveaxis(data, 0, -1)
    a = data[:, 10:42, 10:42]
    b = data[:, 10:42, 10:42].copy()
    c = numpy.empty(a.shape, 'float32')
    d = numpy.empty(a.shape, 'float32')
    with timer('rfftnd_simulation_channel'):
        with rfftnd(a.shape, mode=API.AXIS0 | API.MODE_FCS) as fft:
            fft.crosscorrelate(a, b, out=c)
            fft.autocorrelate(a, out=d)
    assert numpy.allclose(c, d)


def test_yxt_ipcf():
    """Test yxt_ipcf function."""
    data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    data.shape = -1, 64, 64  # txy
    data = numpy.moveaxis(data, 0, -1)  # yxt
    data //= 2
    data = data.astype('int16')

    with timer('yxt_ipcf'):
        result, bins, points = yxt_ipcf(
            data, radius=4, nbins=32, smooth=0.7, nthreads=0
        )

    # fmt: off
    assert_array_equal(
        bins,
        [1, 2, 3, 4, 5, 7, 10, 13, 18, 24, 32, 43, 58, 78, 104, 139, 187, 250,
         334, 447, 598, 800, 1070, 1432, 1915, 2561, 3425, 4580, 6125, 8192]
    )
    assert_array_equal(
        points.flat,
        [4, 0, 4, 1, 3, 2, 3, 3, 2, 3, 1, 4, 0, 4, -1, 4, -2, 3, -3, 3, -3,
         2, -4, 1, -4, 0, -4, -1, -3, -2, -3, -3, -2, -3, -1, -4, 0, -4, 1,
         -4, 2, -3, 3, -3, 3, -2, 4, -1]
    )
    # fmt: on

    known = numpy.fromfile(
        HERE + 'Simulation_Channel.ipcf.bin', dtype='float32'
    )
    known.shape = result.shape
    assert_allclose(result, known, atol=1e-6)


def test_yxt_apcf():
    """Test yxt_apcf function."""
    data = numpy.fromfile(HERE + 'Airy_Detectors.bin', dtype='uint16')

    # correct_bleaching needs 3D
    data.shape = 1, 32, 10000000
    with timer('yxt_correct_bleaching'):
        mean = yxt_correct_bleaching(data, smooth=0.99, nthreads=0)
    known = numpy.fromfile(HERE + 'Airy_Detectors.mean.bin', dtype='float64')
    known.shape = mean.shape
    assert_allclose(mean, known, atol=1e-6)

    # apcf needs 2D
    data.shape = data.shape[1:]
    with timer('yxt_apcf'):
        result, bins = yxt_apcf(data, nbins=256, smooth=0.7)
    known = numpy.fromfile(HERE + 'Airy_Detectors.apcf.bin', dtype='float32')
    known.shape = result.shape
    assert_allclose(result, known, atol=1e-6)
    # fmt: off
    assert_array_equal(
        bins,
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21,
         22, 23, 25, 26, 28, 30, 32, 34, 36, 38, 40, 43, 45, 48, 51, 54, 58,
         61, 65, 69, 74, 78, 83, 88, 94, 99, 106, 112, 119, 126, 134, 143, 151,
         161, 171, 181, 192, 204, 217, 230, 245, 260, 276, 293, 311, 330, 350,
         372, 395, 419, 445, 473, 502, 533, 566, 601, 638, 677, 719, 763, 810,
         860, 913, 969, 1029, 1093, 1160, 1231, 1307, 1388, 1473, 1564, 1661,
         1763, 1872, 1987, 2110, 2240, 2378, 2524, 2680, 2845, 3020, 3207,
         3404, 3614, 3837, 4073, 4324, 4591, 4874, 5174, 5493, 5832, 6191,
         6573, 6978, 7408, 7864, 8349, 8863, 9410, 9990, 10605, 11259, 11953,
         12689, 13471, 14301, 15183, 16118, 17112, 18166, 19286, 20474, 21736,
         23076, 24498, 26007, 27610, 29312, 31118, 33036, 35072, 37233, 39528,
         41963, 44549, 47295, 50209, 53304, 56589, 60076, 63778, 67708, 71881,
         76311, 81013, 86006, 91306, 96933, 102906, 109248, 115981, 123128,
         130716, 138771, 147323, 156402, 166040, 176272, 187135, 198668,
         210910, 223908, 237706, 252355, 267906, 284416, 301944, 320551,
         340305, 361276, 383540, 407176, 432268, 458907, 487187, 517210,
         549083, 582921, 618843, 656980, 697466, 740448, 786078, 834520,
         885948, 940544, 998506, 1060039, 1125364, 1194715, 1268339, 1346501,
         1429479, 1517571, 1611092, 1710376, 1815778, 1927676, 2046469,
         2172583, 2306469, 2448606, 2599502, 2759696, 2929763, 3110311,
         3301984, 3505470, 3721495, 3950833, 4194304]
    )
    # fmt: on


def test_yxt_lstics():
    """Test yxt_lstics function."""
    data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    data.shape = -1, 64, 64  # txy
    data = numpy.moveaxis(data, 0, -1)  # yxt

    with timer('yxt_subtract_immobile'):
        yxt_subtract_immobile(data)

    with timer('yxt_lstics'):
        result, bins, lines = yxt_lstics(
            data,
            block=(16, 16, 2, 2),
            nlines=16,
            linelength=8,
            nbins=16,
            smooth=0.0,
        )

    known = numpy.fromfile(
        HERE + 'Simulation_Channel.lstics.bin', dtype='float32'
    )
    known.shape = result.shape
    assert_allclose(result, known, atol=1e-6)
    assert_array_equal(
        bins,
        [1, 3, 6, 11, 20, 36, 67, 122, 222, 406, 741, 1351, 2463, 4492, 8192],
    )
    # fmt: off
    assert_array_equal(
        lines,
        [[[0, 0], [1, 0], [2, 0], [3, 0], [4, 0], [5, 0], [6, 0], [7, 0]],
         [[0, 0], [1, 0], [2, 1], [3, 1], [4, 2], [5, 2], [6, 2], [6, 3]],
         [[0, 0], [1, 1], [1, 1], [2, 2], [3, 3], [4, 4], [4, 4], [5, 5]],
         [[0, 0], [0, 1], [1, 2], [1, 3], [2, 4], [2, 5], [2, 6], [3, 6]],
         [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7]],
         [[0, 0], [0, 1], [-1, 2], [-1, 3],
          [-2, 4], [-2, 5], [-2, 6], [-3, 6]],
         [[0, 0], [-1, 1], [-1, 1], [-2, 2],
          [-3, 3], [-4, 4], [-4, 4], [-5, 5]],
         [[0, 0], [-1, 0], [-2, 1], [-3, 1],
          [-4, 2], [-5, 2], [-6, 2], [-6, 3]],
         [[0, 0], [-1, 0], [-2, 0], [-3, 0],
          [-4, 0], [-5, 0], [-6, 0], [-7, 0]],
         [[0, 0], [-1, 0], [-2, -1], [-3, -1],
          [-4, -2], [-5, -2], [-6, -2], [-6, -3]],
         [[0, 0], [-1, -1], [-1, -1], [-2, -2],
          [-3, -3], [-4, -4], [-4, -4], [-5, -5]],
         [[0, 0], [0, -1], [-1, -2], [-1, -3],
          [-2, -4], [-2, -5], [-2, -6], [-3, -6]],
         [[0, 0], [0, -1], [0, -2], [0, -3],
          [0, -4], [0, -5], [0, -6], [0, -7]],
         [[0, 0], [0, -1], [1, -2], [1, -3],
          [2, -4], [2, -5], [2, -6], [3, -6]],
         [[0, 0], [1, -1], [1, -1], [2, -2],
          [3, -3], [4, -4], [4, -4], [5, -5]],
         [[0, 0], [1, 0], [2, -1], [3, -1],
          [4, -2], [5, -2], [6, -2], [6, -3]]],
    )
    # fmt: on


def test_yxt_imsd():
    """Test yxt_imsd function."""
    data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    data.shape = -1, 64, 64  # txy
    data = numpy.moveaxis(data, 0, -1)  # yxt

    with timer('yxt_subtract_immobile'):
        yxt_subtract_immobile(data)

    with timer('yxt_imsd'):
        result = yxt_imsd(data, block=(32, 32, 4, 4), bins=16, smooth=0.0)

    result = numpy.moveaxis(result, -1, -3)  # SimFCS axes order
    known = numpy.fromfile(
        HERE + 'Simulation_Channel.imsd.bin', dtype='float32'
    )
    known.shape = result.shape
    assert_allclose(result, known, atol=1e-6)


def test_yxt_subtract_immobile(dtype='int16', nthreads=0):
    """Test yxt_subtract_immobile function."""
    data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    data.shape = -1, 64, 64  # txy
    if dtype == 'int16':
        data //= 2
        data = data.astype(dtype)

    def subtract_immobile_np(a):
        dt = a.dtype
        a = numpy.ascontiguousarray(a).astype('float64')
        im = numpy.mean(a, axis=2)
        im -= numpy.mean(im)
        a -= im[..., numpy.newaxis]
        a = numpy.round(a)
        a = numpy.clip(a, numpy.iinfo(dt).min, numpy.iinfo(dt).max)
        return a.astype(dt)

    with timer('subtract_immobile_np'):
        corrected_np = subtract_immobile_np(numpy.moveaxis(data, 0, -1))

    # non-contiguous
    corrected = data.copy()
    corrected = numpy.moveaxis(corrected, 0, -1)  # need yxt
    # print(corrected.shape, corrected.strides)
    with timer('yxt_subtract_immobile tyx'):
        yxt_subtract_immobile(corrected, nthreads=nthreads)
    assert numpy.allclose(corrected, corrected_np)

    # contiguous
    corrected = data.copy()
    corrected = numpy.moveaxis(corrected, 0, -1)  # need yxt
    corrected = numpy.ascontiguousarray(corrected)
    # print(corrected.shape, corrected.strides)
    with timer('yxt_subtract_immobile yxt'):
        yxt_subtract_immobile(corrected, nthreads=nthreads)

    assert numpy.allclose(corrected, corrected_np)


def test_yxt_correct_bleaching(dtype='int16', smooth=0.99, nthreads=0):
    """Test yxt_correct_bleaching function."""
    # data = numpy.fromfile(HERE + 'Simulation_Channel.bin', dtype='uint16')
    # data.shape = -1, 64, 64  # txy
    data = numpy.fromfile(HERE + 'bleach.bin', 'uint16').reshape((64, 64, -1))
    data = numpy.moveaxis(data, -1, 0).copy()
    if dtype == 'int16':
        data //= 2
        data = data.astype(dtype)

    def correct_bleaching_np(data, smooth=smooth):
        f0 = smooth
        f1 = 1.0 - smooth
        dt = data.dtype
        a = numpy.ascontiguousarray(data).astype('float64')
        mean = numpy.mean(a, axis=2)
        for i in range(1, a.shape[2]):
            a[..., i] = data[..., i] * f1 + a[..., i - 1] * f0
        for i in range(a.shape[2] - 2, -1, -1):
            a[..., i] = a[..., i] * f1 + a[..., i + 1] * f0

        deficit = numpy.reciprocal(a)
        deficit *= mean[..., None]
        deficit = numpy.sqrt(numpy.abs(deficit))
        a *= -1.0
        a += data
        a *= deficit
        a += mean[..., None]
        a = numpy.round(a)
        a = numpy.clip(a, numpy.iinfo(dt).min, numpy.iinfo(dt).max)
        return a.astype(dt), mean

    with timer('correct_bleaching_np'):
        corrected_np, mean_np = correct_bleaching_np(
            numpy.moveaxis(data, 0, -1)
        )

    # non-contiguous
    corrected = data.copy()
    corrected = numpy.moveaxis(corrected, 0, -1)  # need yxt
    # print(corrected.shape, corrected.strides)
    with timer('yxt_correct_bleaching tyx'):
        mean = yxt_correct_bleaching(corrected, smooth, nthreads=nthreads)

    if 0:
        pyplot.imshow(mean)  # type: ignore[unreachable]
        pyplot.figure()
        pyplot.plot(data[:, 10, 10])
        pyplot.plot(corrected[10, 10])
        pyplot.show()

    assert numpy.allclose(corrected, corrected_np)
    assert numpy.allclose(mean, mean_np)

    # contiguous
    corrected = data.copy()
    corrected = numpy.moveaxis(corrected, 0, -1)  # need yxt
    corrected = numpy.ascontiguousarray(corrected)
    # print(corrected.shape, corrected.strides)
    with timer('yxt_correct_bleaching yxt'):
        mean = yxt_correct_bleaching(corrected, smooth, nthreads=nthreads)

    assert numpy.allclose(corrected, corrected_np)
    assert numpy.allclose(mean, mean_np)


@pytest.mark.xfail(reason='results depend on compiler version, MKL version')
def test_ipcf_nlsp_1dpcf():
    """Test ipcf_nlsp_1dpcf function."""
    ipcf = numpy.fromfile(
        HERE + 'Simulation_Channel.ipcf.bin', dtype='float32'
    )
    ipcf.shape = 56, 56, 24, 30

    bins = logbins(8192, 32)
    times = bins2times(bins, frametime=0.01)
    assert len(times) == ipcf.shape[3]
    points = circle(4)
    distances = points2distances(points, pixelsize=0.05)
    assert len(distances) == ipcf.shape[2]

    args = [0.1, 1.0, 0.3 * 0.3]
    bounds = [1e-6, 1e3, 1e-6, 1e3]
    # fmt: off
    settings = [
        400.0, 200.0, 1e-10, 1e-10, 1e-10, 1e-10, 1e-10, 1e-10, 1e-10, 0.0
    ]
    # fmt: on

    with timer('ipcf_nlsp_1dpcf'):
        x, fx, status = ipcf_nlsp_1dpcf(
            ipcf,
            times,
            distances,
            args,
            bounds=bounds,
            ix=None,
            ifx=None,
            status=None,
            settings=settings,
            average=False,
            nthreads=0,
        )

    ipcf_x = numpy.fromfile(
        HERE + 'Simulation_Channel.ipcf_x.bin', dtype='float32'
    )
    ipcf_x.shape = x.shape
    assert_allclose(x, ipcf_x)

    ipcf_fx = numpy.fromfile(
        HERE + 'Simulation_Channel.ipcf_fx.bin', dtype='float32'
    )
    ipcf_fx.shape = fx.shape
    assert_allclose(fx[..., 1:-1], ipcf_fx[..., 1:-1])

    ipcf_status = numpy.fromfile(
        HERE + 'Simulation_Channel.ipcf_status.bin', dtype='float32'
    )
    ipcf_status.shape = status.shape
    # assert_allclose(ipcf_status, status)


def test_yxt_dft():
    """Test yxt_dft function."""
    data = numpy.fromfile(HERE + 'LifetimeHistogram.bin', dtype='float32')
    data.shape = 256, 256, 256

    with timer('yxt_dft'):
        result = yxt_dft(numpy.moveaxis(data, 0, -1), samples=5)

    with timer('yxt_dft numpy'):
        expected = numpy.fft.fft(data, axis=0).astype('complex64')

    atol = 1e-6
    assert_allclose(result[0], expected[0].real, atol=atol)
    assert_allclose(result[1], expected[1].real, atol=atol)
    assert_allclose(result[2], expected[1].imag, atol=atol)
    assert_allclose(result[3], expected[2].real, atol=atol)
    assert_allclose(result[4], expected[2].imag, atol=atol)


@contextmanager
def timer(message):
    """Context manager for timing execution speed of body."""
    print(message, end=' ')
    start = time.time()
    yield
    print(f'took {1000 * (time.time() - start):1.0f} ms')


def runall():
    """Run all tests."""
    import inspect

    for func in (
        obj
        for name, obj in inspect.getmembers(sys.modules[__name__])
        if inspect.isfunction(obj) and name.startswith('test_')
    ):
        try:
            func()
        except NotImplementedError:
            pass
    print('All done.')


if __name__ == '__main__':
    if os.environ.get('WINGDB_ACTIVE') or os.environ.get('TERM_PROGRAM'):
        # running in IDE
        sys.path = ['..', *sys.path]

        from matplotlib import pyplot

        import icsdll
        from icsdll import *  # noqa: F403

        runall()
    else:
        # run pytests
        import warnings

        import icsdll
        from icsdll import *  # noqa: F403

        # warnings.simplefilter('always')
        warnings.filterwarnings('ignore', category=ImportWarning)
        argv = sys.argv
        # argv.append('--cov=icsdll')
        argv.append('--verbose')
        sys.exit(pytest.main(argv))
else:
    # pytest
    import icsdll
    from icsdll import (
        API,
        bins2times,
        circle,
        ipcf_nlsp_1dpcf,
        logbins,
        nextpow2,
        numpy_correlate,
        numpy_correlate1d,
        points2distances,
        radial,
        rfftnd,
        yxt_apcf,
        yxt_correct_bleaching,
        yxt_dft,
        yxt_imsd,
        yxt_ipcf,
        yxt_lstics,
        yxt_subtract_immobile,
    )


# mypy: allow-untyped-defs, allow-untyped-calls
# mypy: disable-error-code="attr-defined"



# icsdll/setup.py

"""Icsdll package setuptools script."""

import glob
import os
import re
import sys

from setuptools import setup
from setuptools._distutils import ccompiler

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):  # type: ignore[no-redef]
        """Override wheel tag to py3-none-win_amd64."""

        def get_tag(self) -> tuple[str, str, str]:
            return 'py3', 'none', 'win_amd64'

    cmdclass: dict[str, object] = {'bdist_wheel': bdist_wheel}
except ImportError:
    cmdclass = {}

ONEAPI_ROOT = os.environ.get('ONEAPI_ROOT', r'C:\PROGRA~2\Intel\oneAPI')


def search(pattern: str, string: str, flags: int = 0) -> str:
    """Return first match of pattern in string."""
    match = re.search(pattern, string, flags)
    if match is None:
        raise ValueError(f'{pattern!r} not found')
    return match.groups()[0]


def fix_docstring_examples(docstring: str) -> str:
    """Return docstring with examples fixed for GitHub."""
    start = True
    indent = False
    lines = ['..', '  This file is generated by setup.py', '']
    for line in docstring.splitlines():
        if not line.strip():
            start = True
            indent = False
        if line.startswith('>>> '):
            indent = True
            if start:
                lines.extend(['.. code-block:: python', ''])
                start = False
        lines.append(('    ' if indent else '') + line)
    return '\n'.join(lines)


with open('icsdll/icsdll.py', encoding='utf-8') as fh:
    code = fh.read()

version = search(r"__version__ = '(.*?)'", code).replace('.x.x', '.dev0')

description = search(r'"""(.*)\.(?:\r\n|\r|\n)', code)

readme = search(
    r'(?:\r\n|\r|\n){2}"""(.*)"""(?:\r\n|\r|\n){2}from __future__',
    code,
    re.MULTILINE | re.DOTALL,
)
readme = '\n'.join(
    [description, '=' * len(description), *readme.splitlines()[1:]]
)


kwargs: dict[str, object] = {}

if 'sdist' in sys.argv:
    # update LICENSE and README files

    with open('README.rst', 'w', encoding='utf-8') as fh:
        fh.write(fix_docstring_examples(readme))

    license = search(
        r'(# Copyright.*?(?:\r\n|\r|\n))(?:\r\n|\r|\n)+""',
        code,
        re.MULTILINE | re.DOTALL,
    )
    license = license.replace('# ', '').replace('#', '')

    with open('LICENSE', 'w', encoding='utf-8') as fh:
        fh.write('BSD-3-Clause license\n\n')
        fh.write(license)

elif (
    sys.platform == 'win32'
    and '64 bit' in sys.version
    and sys.version_info[:2] >= (3, 11)
):
    # build DLL for 64-bit Windows
    for f in glob.glob(os.path.join('icsdll', 'ICSx64.*')):
        try:
            os.remove(f)
        except Exception:  # noqa: S110
            pass

    sources = [
        os.path.join('ics', c)
        for c in ('dllmain.cpp', 'ics.cpp', 'rfft.cpp', 'yxt.cpp', 'nlsp.cpp')
    ]

    compiler = ccompiler.new_compiler()
    objects = compiler.compile(
        sources,
        'build',
        extra_preargs=[
            '/EHsc',
            '/FD',
            '/DICS_EXPORTS',
            '/DICS_SIMFCS',
            '/openmp',
        ],
        include_dirs=['ics', ONEAPI_ROOT + '/mkl/latest/include'],
    )

    compiler.link_shared_lib(
        objects,
        os.path.join('icsdll', 'ICSx64'),
        extra_preargs=['/DLL'],
        libraries=[
            'mkl_core',
            'mkl_intel_lp64',
            'mkl_sequential',
            # 'mkl_intel_thread',
            # 'libiomp5md',
        ],
        library_dirs=[ONEAPI_ROOT + '/mkl/latest/lib'],
    )
    kwargs = {
        'package_data': {'icsdll': ['ICSx64.dll', '*.dll']},
        'libraries': [('', {'sources': []})],  # sets ispurelib = False
    }
else:
    kwargs = {}


setup(
    name='icsdll',
    cmdclass=cmdclass,
    version=version,
    description=description,
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Christoph Gohlke',
    author_email='cgohlke@cgohlke.com',
    url='https://www.cgohlke.com',
    project_urls={
        'Bug Tracker': 'https://github.com/cgohlke/icsdll/issues',
        'Source Code': 'https://github.com/cgohlke/icsdll',
        # 'Documentation': 'https://',
    },
    python_requires='>=3.11',
    install_requires=['numpy'],
    packages=['icsdll'],
    license='BSD-3-Clause',
    zip_safe=False,
    platforms=['Windows'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
    ],
    **kwargs,
)
